# shapes
RECT = 'rect' #:
SQUARE = 'rect'
CIRCLE = 'circle' #:
ELLIPSE = 'ellipse' #:

# colors
BLACK = (0, 0, 0) #:
SILVER = (192, 192, 192) #:
GRAY = (128, 128, 128) #:
WHITE = (255, 255, 255) #:
RED = (255, 0, 0) #:
MAROON = (128, 0, 0) #:
ORANGE = (255, 165, 0) #:
YELLOW = (255, 255, 0) #:
OLIVE = (128, 128, 0) #:
LIME = (0, 255, 0) #:
GREEN = (0, 128, 0) #:
AQUA = (0, 255, 255) #:
TEAL = (0, 128, 128) #:
BLUE = (0, 0, 255) #:
NAVY = (0, 0, 128) #:
CYAN = (0, 200, 200) #:
PINK = (255, 0, 255) #:
PURPLE = (128, 0, 128) #:

# Actor actions (omnidirectional)
IDLE = 'idle'
JUMP = 'jump'
RUN = 'run'
WALK = 'walk'
THROW = 'throw'
HURT = 'hurt'
DIE = 'die'
ATTACK = 'attack'
SHOOT = 'shoot'
HAPPY = 'happy'
DANCE = 'dance'

# Actor actions (directional)
ATTACK_BACK = 'attack_back'
ATTACK_FRONT = 'attack_front'
ATTACK_LEFT = 'attack_left'
ATTACK_RIGHT = 'attack_right'

SHOOT_BACK = 'shoot_back'
SHOOT_FRONT = 'shoot_front'
SHOOT_LEFT = 'shoot_left'
SHOOT_RIGHT = 'shoot_right'

THROW_BACK = 'throw_back'
THROW_FRONT = 'throw_front'
THROW_LEFT = 'throw_left'
THROW_RIGHT = 'throw_right'

DIE_BACK = 'die_back'
DIE_FRONT = 'die_front'
DIE_LEFT = 'die_left'
DIE_RIGHT = 'die_right'

FLY_BACK = 'fly_back'
FLY_FRONT = 'fly_front'
FLY_LEFT = 'fly_left'
FLY_RIGHT = 'fly_right'

JUMP_BACK = 'jump_back'
JUMP_FRONT = 'jump_front'
JUMP_LEFT = 'jump_left'
JUMP_RIGHT = 'jump_right'

IDLE_BACK = 'idle_back'
IDLE_FRONT = 'idle_front'
IDLE_LEFT = 'idle_left'
IDLE_RIGHT = 'idle_right'

IDLE_AIM_BACK = 'idle_aim_back'
IDLE_AIM_FRONT = 'idle_aim_front'
IDLE_AIM_LEFT = 'idle_aim_left'
IDLE_AIM_RIGHT = 'idle_aim_right'

WALK_BACK = 'walk_back'
WALK_FRONT = 'walk_front'
WALK_LEFT = 'walk_left'
WALK_RIGHT = 'walk_right'

WALK_AIM_BACK = 'walk_aim_back'
WALK_AIM_FRONT = 'walk_aim_front'
WALK_AIM_LEFT = 'walk_aim_left'
WALK_AIM_RIGHT = 'walk_aim_right'

RUN_BACK = 'run_back'
RUN_FRONT = 'run_front'
RUN_LEFT = 'run_left'
RUN_RIGHT = 'run_right'

DANCE_BACK = 'dance_back'
DANCE_FRONT = 'dance_front'
DANCE_LEFT = 'dance_left'
DANCE_RIGHT = 'dance_right'

LEFT = 'left'
RIGHT = 'right'
BACK = 'back'
FRONT = 'front'
POSE = 'pose'

OBSTACLE = 'wall'
GRAVITY = 0

# other contants
FOREVER = -1

# positions for scoring
UPPER_LEFT = 0
UPPER_RIGHT = 1
LOWER_LEFT = 2
LOWER_RIGHT = 3

ACCUMULATE = 0
TIMER = 1
VALUE = 2

WIN = 1
LOSE = 2
